
import argparse
import coopr.misc.coopr_parser
import os.path
import textwrap
import coopr.opt

def setup_solvers_parser(parser):
    parser.add_argument("--list", dest="summary", action='store_true', default=False,
                        help="list the active solvers")
    parser.add_argument("--options", dest="options", action='store_true', default=False,
                        help="print the solver options that are supported by solvers")

def setup_test_parser(parser):
    parser.add_argument('--csv-file', '--csv', action='store', dest='csv', default=None,
                        help='Save test results to this file in a CSV format')
    parser.add_argument("-d", "--debug", action="store_true", dest="debug", default=False,
                        help="Show debugging information and text generated during tests.")
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose", default=False,
                        help="Show verbose results output.")
    parser.add_argument("solver", metavar="SOLVER", default=None, nargs='*',
                        help="a solver name")

def print_solvers():
    wrapper = textwrap.TextWrapper(replace_whitespace=False)
    print wrapper.fill("The following pre-defined solver interfaces are recognized by Pyomo:")
    print ""
    solver_list = coopr.opt.SolverFactory.services()
    solver_list = sorted( filter(lambda x: '_' != x[0], solver_list) )
    n = max(map(len, solver_list))
    wrapper = textwrap.TextWrapper(subsequent_indent=' '*(n+9))
    for s in solver_list:
        format = '    %-'+str(n)+'s  %s'
        print wrapper.fill(format % (s , coopr.opt.SolverFactory.doc(s)))
    print ""
    wrapper = textwrap.TextWrapper(subsequent_indent='')
    print wrapper.fill('The default solver is glpk.')
    print ""
    print wrapper.fill('Subsolver options can be specified by with the solver name followed by colon and then the subsolver.  For example, the following specifies that the asl solver will be used:')
    print '   --asl:PICO'
    print wrapper.fill('This indicates that the asl solver will launch the PICO executable to perform optimization.  Currently, no other solver supports this syntax.')


def main_exec(options):
    import coopr.pyomo.check as check
    if options.options:
        print "TODO - print solver options"
    elif options.summary:
        print_solvers()
    else:
        print "No solver action specified."

def test_exec(options):
    try:
        import coopr.data.pyomo
    except ImportError:
        print "Cannot test solvers.  The package coopr.data.pyomo is not installed!"
        return
    coopr.data.pyomo.test_solvers(options)
    
    
#
# Add a subparser for the coopr command
#
setup_solvers_parser(
    coopr.misc.coopr_parser.add_subparser('solvers',
        func=main_exec, 
        help='Print information on Coopr solvers.',
        description='This coopr subcommand is used to print solver information.',
        epilog='Note that the different options are meant to be used exclusively.  Additionally, the solver name is not used when listing all solvers.'
        ))

setup_test_parser(
    coopr.misc.coopr_parser.add_subparser('test-solvers',
        func=test_exec,
        help='Test Coopr solvers',
        description='This coopr subcommand is used to run tests on installed solvers.',
        epilog="""
This Coopr subcommand executes solvers on a variety of test problems that
are defined in the coopr.data.pyomo package.  The default behavior is to
test all available solvers, but the testing can be limited by explicitly
specifying the solvers that are tested.  For example:

  coopr test-solvers glpk cplex

will test only the glpk and cplex solvers.

The configuration file test_solvers.yml in coopr.data.pyomo defines a
series of test suites, each of which specifies a list of solvers that are
tested with a list of problems.  For each solver-problem pair, the Pyomo
problem is created and optimized with the the Coopr solver interface.
The optimization results are then analyzed using a function with the
same name as the test suite (found in the coopr/data/pyomo/plugins
directory).  These functions perform a sequence of checks that compare
the optimization results with baseline data, evaluate the solver return
status, and otherwise verify expected solver behavior.

The default summary is a simple table that describes the percentage of
checks that passed.  The '-v' option can be used to provide a summary
of all checks that failed, which is generally useful for evaluating
solvers.  The '-d' option provides additional detail about all checks
performed (both passed and failed checks).  Additionally, this option
prints information about the optimization process, such as the pyomo
command-line that was executed.""",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
)


def main(args=None):
    parser = argparse.ArgumentParser()
    setup_parser(parser)
    parser.set_defaults(func=main_exec)
    ret = parser.parse_args(args)
    ret = ret.func(ret)




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['SolverResults']

import math
import sys
from container import *
from pyutilib.enum import Enum
from pyutilib.misc import Bunch
import copy

import StringIO

import problem
import solver
import solution

import json
try:
    import yaml
    yaml_available=True
except ImportError:
    yaml_available=False


class SolverResults(MapContainer):

    undefined = undefined
    default_print_options = solution.default_print_options

    # XXX begin debugging
    def __del__(self):
        MapContainer.__del__(self)
        #print("Called __del__ on SolverResults")
        self._sections = None
        self._descriptions = None
        self._symbol_map = None
    # XXX end debugging

    def __init__(self):
        MapContainer.__init__(self)
        self._sections = []
        self._symbol_map = None
        self._descriptions = {}
        self.add('problem', ListContainer(problem.ProblemInformation), False, "Problem Information")
        self.add('solver', ListContainer(solver.SolverInformation), False, "Solver Information")
        self.add('solution', solution.SolutionSet(), False, "Solution Information")

    def __getstate__(self):
        def _canonical_label(obj):
            if obj is obj.component():
                label = obj.name
            else:
                index = obj.index
                if type(index) is not tuple:
                    index = (index,)
                codedIdx = []
                for idx in index:
                    if idx is None:
                        codedIdx.append('!')
                    elif type(idx) is str:
                        codedIdx.append('$'+idx)
                    elif int(idx) == idx:
                        codedIdx.append('#'+str(idx))
                    else:
                        raise ValueError(
                            "Unexpected type %s encountered when pickling "
                            "SolverResults object index: %s" %
                            (str(type(idx)), str(obj.index)))
                obj = obj.component()
                label = obj.name + ':' + ','.join(codedIdx)
            if obj.parent is None or obj.parent() is None:
                return label
            obj = obj.parent()
            while obj.parent is not None and obj.parent() is not None:
                label = str(obj.name) + '.' + label
                obj = obj.parent()
            return label
            
        sMap = self._symbol_map
        if sMap is None:
            return MapContainer.__getstate__(self)
        for soln in self.solution:
            for symbol, obj in soln.objective.iteritems():
                obj.canonical_label = _canonical_label(sMap.getObject(symbol))
            for symbol, var in soln.variable.iteritems():
                if symbol == 'ONE_VAR_CONSTANT':
                    continue
                var['canonical_label'] = _canonical_label(sMap.getObject(symbol))
            for symbol, con in soln.constraint.iteritems():
                con.canonical_label = _canonical_label(sMap.getObject(symbol))
        results = MapContainer.__getstate__(self)
        results['_symbol_map'] = None
        return results

    def add(self, name, value, active, description):
        self.declare(name, value=value, active=active)
        tmp = self._convert(name)
        self._sections.append(tmp)
        self._descriptions[tmp]=description

    def json_repn(self, options=None):
        if options is None:
            return self._repn_(SolverResults.default_print_options)
        else:
            return self._repn_(options)

    def _repn_(self, option):
        if not option.schema and not self._active and not self._required:
            return ignore
        tmp = {}
        for key in self._sections:
            rep = dict.__getitem__(self, key)._repn_(option)
            if not rep == ignore:
                tmp[key] = rep
        return tmp

    def write(self, **kwds):
        if 'filename' in kwds:
            OUTPUT=open(kwds['filename'],"w")
            del kwds['filename']
            kwds['ostream']=OUTPUT
            self.write(**kwds)
            OUTPUT.close()
            return

        if not 'format' in kwds or kwds['format'] == 'yaml':
            self.write_yaml(**kwds)
            return
        #
        # Else, write in JSON format
        #
        repn = self.json_repn()
        if 'ostream' in kwds:
            ostream = kwds['ostream']
            del kwds['ostream']
        else:
            ostream = sys.stdout
        for soln in repn.get('Solution', []):
            for data in ['Variable', 'Constraint', 'Objective']:
                remove = set()
                data_value = soln.get(data,{})
                if not isinstance(data_value,dict):
                    continue
                for kk,vv in soln.get(data,{}).iteritems():
                    tmp = {}
                    for k,v in vv.iteritems():
                        if k == 'Id' or (k != 'Id' and math.fabs(v) != 0.0):
                            tmp[k] = v
                    if len(tmp) > 1 or (len(tmp) == 1 and not 'Id' in tmp):
                        soln[data][kk] = tmp
                    else:
                        remove.add((data,kk))
                for item in remove:
                    del soln[item[0]][item[1]]
        json.dump(repn, ostream, indent=4, sort_keys=True)

    def write_yaml(self, **kwds):
        if 'ostream' in kwds:
            ostream = kwds['ostream']
            del kwds['ostream']
        else:
            ostream = sys.stdout

        option = copy.copy(SolverResults.default_print_options)
        for key in kwds:
            setattr(option,key,kwds[key])

        repn = self._repn_(option)
        print >>ostream, "# =========================================================="
        print >>ostream, "# = Solver Results                                         ="
        print >>ostream, "# =========================================================="
        for i in xrange(len(self._order)):
            key = self._order[i]
            if not key in repn:
                continue
            item = dict.__getitem__(self,key)
            print >>ostream, ""
            print >>ostream, "# ----------------------------------------------------------"
            print >>ostream, "#   %s" % self._descriptions[key]
            print >>ostream, "# ----------------------------------------------------------"
            print >>ostream, key+":",
            if isinstance(item, ListContainer):
                item.pprint(ostream, option, prefix="", repn=repn[key])
            else:
                item.pprint(ostream, option, prefix="  ", repn=repn[key])

    def read(self, **kwds):
        if 'istream' in kwds:
            istream = kwds['istream']
            del kwds['istream']
        else:
            ostream = sys.stdin
        if 'filename' in kwds:
            INPUT=open(kwds['filename'],"r")
            del kwds['filename']
            kwds['istream']=INPUT
            self.read(**kwds)
            INPUT.close()
            return

        if not 'format' in kwds or kwds['format'] == 'yaml':
            if not yaml_available:
                raise IOError, "Aborting SolverResults.read() because PyYAML is not installed!"
            repn = yaml.load(istream, Loader=yaml.SafeLoader)
        else:
            repn = json.load(istream)
        for i in xrange(len(self._order)):
            key = self._order[i]
            if not key in repn:
                continue
            item = dict.__getitem__(self,key)
            item.load(repn[key])

    def __repr__(self):
        return str(self._repn_(SolverResults.default_print_options))

    def __str__(self):
        ostream = StringIO.StringIO()
        option=SolverResults.default_print_options
        self.pprint(ostream, option, repn=self._repn_(option))
        return ostream.getvalue()


if __name__ == '__main__':
    results = SolverResults()
    results.write(schema=True)
    #print results



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = [ 'SymbolMap', 'symbol_map_from_instance' ]

# 
# an experimental utility method to create a symbol map from an instance. really will 
# only work if name-based labelers are used, but that's good enough for now. also 
# won't work with blocks. 
#

def symbol_map_from_instance(instance):

   from coopr.pyomo.base import Var, _VarValue, label_from_name, Constraint, ConstraintData, Objective, ObjectiveData 
   from coopr.pyomo.io.cpxlp import CPXLP_text_labeler

   resulting_map = SymbolMap(instance)

   labeler = CPXLP_text_labeler()

   for block in instance.all_blocks():
      for variable in block.active_components(Var).itervalues():
         for varvalue in variable.itervalues():
            # ignore the return value - we're just trying to populate the map.
            symbol = resulting_map.getSymbol(varvalue, labeler)

   for block in instance.all_blocks():
       active_constraints = block.active_components(Constraint)
       for constraint in active_constraints.itervalues():
           for constraint_data in constraint.itervalues():
               con_symbol = resulting_map.getSymbol( constraint_data, labeler )               
               if constraint_data._equality:               
                   label = 'c_e_' + con_symbol + '_'
                   resulting_map.alias(constraint_data, label)
               else:
                   if constraint_data.lower is not None:
                       label = 'c_l_' + con_symbol + '_'
                       resulting_map.alias(constraint_data, label)
                   if constraint_data.upper is not None:
                       label = 'c_u_' + con_symbol + '_'
                       resulting_map.alias(constraint_data, label)

   for objective in instance.active_components(Objective).itervalues():
      for objective_data in objective.itervalues():
         # ignore the return value - we're just trying to populate the map.
         resulting_map.getSymbol(objective_data, labeler)      
         resulting_map.alias(objective_data, "__default_objective__")

   return resulting_map

#
# a symbol map is a mechanism for tracking assigned labels (e.g., for use when writing 
# problem files for input to an optimizer) for objects in a particular problem instance.
#

class SymbolMap(object):

    class UnknownSymbol:
        pass

    def __init__(self, instance):

        # conceptually, a symbol map must be associated with an instance - for example,
        # the byObject map creates associations between object ids and their symbols
        # => the ids are tied to a specific instance. however, we don't actually do
        # anything with the instance within this class quite yet.
        self.instance = instance
        
        # maps object id()s to their assigned symbol.
        self.byObject = {}

        # maps assigned symbols to the corresponding objects.
        self.bySymbol = {}

        self.aliases = {}

    def getSymbol(self, obj, labeler, *args):

        obj_id = id(obj)
        if obj_id in self.byObject:
           return self.byObject[obj_id]

        # the following test is slightly faster than always calling *args
        if args:
           ans = labeler(obj, *args)
        else:
           ans = labeler(obj)
           
        if self.bySymbol.setdefault(ans, obj) is not obj:
           raise RuntimeError(
              "Duplicate symbol '%s' already associated with "
              "component '%s' (conflicting component: '%s')"
              % (ans, self.bySymbol[ans].name, obj.name) )
        self.byObject[obj_id] = ans
        return ans

    def alias(self, obj, name):

        #if id(obj) not in self.byObject:
        #    raise RuntimeError(
        #        "Cannot alias object '%s': object not in SymbolMap."
        #        % ( name, ))
        if self.aliases.setdefault(name, obj) is not obj:
            raise RuntimeError(
                "Duplicate alias '%s' already associated with "
                "component '%s' (conflicting component: '%s')"
                % (name, self.aliases[name].name, obj.name) )

    def getObject(self, symbol):

        ans = self.bySymbol.get(symbol, SymbolMap.UnknownSymbol)
        if ans is SymbolMap.UnknownSymbol:
            ans = self.aliases.get(symbol, SymbolMap.UnknownSymbol)
        return ans
            
    def getEquivalentObject(self, symbol, instance):

        obj = self.bySymbol.get(symbol, SymbolMap.UnknownSymbol)
        if obj is SymbolMap.UnknownSymbol:
            obj = self.aliases.get(symbol, SymbolMap.UnknownSymbol)
            if obj is SymbolMap.UnknownSymbol:
                return SymbolMap.UnknownSymbol
            
        # FIXME: This is still a hack that needs to be fixed for constraints
        # Actually, it works for constraints, objectives, and vars, but
        # not for other indexable objects.  I am not sure if that is a
        # problem, though,
        if 'parent' in obj.__dict__:
            idx = None
        else:
            idx = obj.index
            obj = obj.component()
        path = []
        while obj is not None:
            path.append(obj)
            obj = obj.parent
        path.pop() # skip the model itself
        obj = instance
        while path:
            try:
                obj = getattr(obj, path.pop().name)
            except AttributeError:
                return SymbolMap.UnknownSymbol
        if idx is not None:
            obj = obj[idx]
        return obj
                
        
        

        
